use std::{collections::HashMap, ops::Add};

use lettre::Transport;
use once_cell::sync::Lazy;
use parking_lot::Mutex;
use ring::rand::SecureRandom;
use rustracing_jaeger::Span;
use tiny_http::{Request, ResponseBox};

use crate::{
    db::{DBConnection, TfaTypes, UserRole},
    dto,
    metrics,
    routes::{auth::tfa::MAIL_SENDER, filters::UserInfo, get_reply, AppError}
};

type RecoveryKeyMap = HashMap<String, (i32, chrono::DateTime<chrono::Utc>)>;
static RECOVERY_KEYS: Lazy<Mutex<RecoveryKeyMap>> = Lazy::new(|| Mutex::new(RecoveryKeyMap::new()));
const KEY_CHARSET: &[u8] = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

pub fn login(
    span: &Span,
    _: &mut Request,
    db: &mut DBConnection,
    data: dto::requests::Login
) -> Result<ResponseBox, AppError> {
    let user = db
        .find_user(span, &data.username, false)
        .ok_or(AppError::Unauthorized("Invalid username or password"))?;

    if !argon2::verify_encoded(user.password.as_str(), data.password.as_bytes()).unwrap_or(false) {
        return AppError::Unauthorized("Invalid username or password").err();
    }

    if user.role == UserRole::Disabled {
        return AppError::Unauthorized("Account is disabled").err();
    }

    if user.tfa_type != TfaTypes::None {
        if let Some(otp) = data.otp {
            if !super::tfa::verify2fa(&user, otp) {
                return AppError::Unauthorized("Incorrect 2fa").err();
            }
        } else {
            if user.tfa_type == TfaTypes::Email {
                super::tfa::send_2fa_mail(span, &user);
            }

            return get_reply(&dto::responses::Success { statusCode: 200 });
        }
    }

    get_reply(&dto::responses::Login {
        statusCode: 200,
        jwt: super::get_token(span, &user, db)
    })
}

pub fn signup(
    span: &Span,
    _: &mut Request,
    db: &mut DBConnection,
    data: dto::requests::SignUp
) -> Result<ResponseBox, AppError> {
    if db.find_user(span, &data.username, false).is_some() {
        return AppError::BadRequest("Username is already taken").err();
    }

    db.create_user_password(span, data.username, super::hash_password(&data.password));

    get_reply(&dto::responses::Success { statusCode: 200 })
}

pub fn send_key(
    span: &Span,
    _: &mut Request,
    db: &mut DBConnection,
    data: dto::requests::SendRecoveryKey
) -> Result<ResponseBox, AppError> {
    // Silently fail to prevent information leak about usernames
    if let Some(user) = db.find_user(span, &data.username, false) {
        let _span = metrics::span("send_recovery_key", span);

        let mut bin = [0_u8; 10];
        crate::routes::auth::SEC_RANDOM.fill(&mut bin).expect("Failed to generate recovery key");
        let code_arr = bin.map(|v| KEY_CHARSET[v as usize % KEY_CHARSET.len()]);
        let code = String::from_utf8_lossy(&code_arr);

        {
            RECOVERY_KEYS.lock().insert(
                code.to_string(),
                (
                    user.id,
                    chrono::Utc::now().add(chrono::Duration::minutes(5))
                )
            );
        }

        let mail = lettre::Message::builder()
            .from("fileserver@mattv.de".parse().unwrap())
            .to(user.name.parse().unwrap())
            .subject("MFileserver - Password reset")
            .body(format!(
                "Your recovery key is: {}\r\nIt is valid for 5 minutes",
                code
            ))
            .unwrap();

        MAIL_SENDER.send(&mail).expect("Failed to send mail");
    }

    get_reply(&dto::responses::Success { statusCode: 200 })
}

pub fn reset(
    span: &Span,
    _: &mut Request,
    db: &mut DBConnection,
    data: dto::requests::ResetPassword
) -> Result<ResponseBox, AppError> {
    if let Some(entry) = { RECOVERY_KEYS.lock().get(&data.key) } {
        if entry.1 < chrono::Utc::now() {
            return AppError::BadRequest("Invalid key").err();
        }
        if let Some(mut user) = db.get_user(span, entry.0) {
            user.password = super::hash_password(&data.password);
            db.save_user(span, &user);
            db.delete_all_tokens(span, user.id);
            get_reply(&dto::responses::Success { statusCode: 200 })
        } else {
            AppError::BadRequest("Invalid user").err()
        }
    } else {
        AppError::BadRequest("Invalid key").err()
    }
}

pub fn refresh(span: &Span, _: &mut Request, db: &mut DBConnection, info: UserInfo) -> Result<ResponseBox, AppError> {
    db.delete_token(span, info.1.id);

    get_reply(&dto::responses::Login {
        statusCode: 200,
        jwt: super::get_token(span, &info.0, db)
    })
}

pub fn logout_all(
    span: &Span,
    _: &mut Request,
    db: &mut DBConnection,
    info: UserInfo
) -> Result<ResponseBox, AppError> {
    db.delete_all_tokens(span, info.0.id);

    get_reply(&dto::responses::Success { statusCode: 200 })
}

pub fn change_password(
    span: &Span,
    _: &mut Request,
    db: &mut DBConnection,
    mut info: UserInfo,
    data: dto::requests::ChangePassword
) -> Result<ResponseBox, AppError> {
    if !argon2::verify_encoded(info.0.password.as_str(), data.oldPassword.as_bytes()).unwrap_or(false) {
        return AppError::Unauthorized("Old password is wrong").err();
    }

    info.0.password = super::hash_password(&data.newPassword);
    db.save_user(span, &info.0);
    db.delete_all_tokens(span, info.0.id);

    get_reply(&dto::responses::Success { statusCode: 200 })
}
